;;; Directory Local Variables
;;; For more information see (info "(emacs) Directory Variables")

((c++-mode
  (flycheck-clang-language-standard . "c++17")
  (flycheck-gcc-language-standard . "c++17")))
